import sys
import logging
import smartcard
import textwrap
import click
from terminaltables import SingleTable
import emv
from emv.card import Card
from emv.protocol.data import Tag, render_element
from emv.protocol.structures import TLV
from emv.protocol.response import ErrorResponse
from emv.exc import InvalidPINException, MissingAppException, CAPError
from emv.util import format_bytes

LOG_LEVELS = {"info": logging.INFO, "debug": logging.DEBUG, "warn": logging.WARN}


def as_table(tlv, title=None, redact=False):
    res = [["Tag", "Name", "Value"]]
    if type(tlv) is not TLV:
        return ""
    for tag, value in tlv.items():
        res.append(
            [
                format_bytes(tag.id),
                tag.name or "",
                "\n".join(textwrap.wrap(render_element(tag, value, redact=redact), 80)),
            ]
        )
    table = SingleTable(res)
    if title is not None:
        table.title = title
    return table.table


def get_reader(reader):
    try:
        return Card(smartcard.System.readers()[reader].createConnection())
    except IndexError:
        click.echo("Reader or card not found")
        sys.exit(2)


def run():
    "Command line entrypoint"
    cli(obj={})


@click.group(
    help="""Utility to interact with EMV payment cards.

Although this tool has been relatively well tested, it's possible to
block or even damage your card, as well as get in trouble with your
card issuer. Please make sure you understand the risks.

Commands marked with [!] will initiate a transaction on the card,
resulting in a permanent change to the card's internal state which
could potentially be detected by your card issuer, particularly if
you initiate many transactions.
"""
)
@click.option(
    "--reader",
    "-r",
    type=int,
    metavar="READER",
    default=0,
    help="the reader to use (default 0)",
)
@click.option(
    "--pin",
    "-p",
    type=str,
    metavar="PIN",
    help="PIN. Note this may be shown in the system process list.",
)
@click.option(
    "--loglevel", "-l", type=str, metavar="LOGLEVEL", default="warn", help="log level"
)
@click.option(
    "--redact/--no-redact",
    default=False,
    help="redact sensitive data for public display. Note that this is not foolproof "
    + """- your card may send sensitive data in tags we don't know about!""",
)
@click.pass_context
def cli(ctx, reader, pin, loglevel, redact):
    logging.basicConfig(level=LOG_LEVELS[loglevel])
    ctx.obj["pin"] = pin
    ctx.obj["reader"] = reader
    ctx.obj["redact"] = redact


@cli.command(help="Show the version of emvtool.")
def version():
    click.echo(emv.__version__)


@cli.command(help="List available card readers.")
def readers():
    click.echo("Available card readers:\n")
    readers = smartcard.System.readers()
    for i in range(0, len(readers)):
        click.echo("%i: %s" % (i, readers[i]))


def render_app(card, df, redact):
    data = card.select_application(df).data

    click.echo(
        as_table(data[Tag.FCI][Tag.FCI_PROP], "FCI Proprietary Data", redact=redact)
    )
    for i in range(1, 31):
        try:
            for j in range(1, 16):
                rec = card.read_record(j, sfi=i).data
                if Tag.RECORD not in rec:
                    break
                click.echo(
                    as_table(rec[Tag.RECORD], "File: %s,%s" % (i, j), redact=redact)
                )
        except ErrorResponse:
            continue


@cli.command(help="Dump card information.")
@click.pass_context
def info(ctx):
    redact = ctx.obj["redact"]
    card = get_reader(ctx.obj["reader"])
    apps = card.list_applications()

    click.secho("\n1PAY.SYS.DDF01 (Index of apps for chip payments)", bold=True)
    try:
        render_app(card, "1PAY.SYS.DDF01", redact)
    except MissingAppException:
        click.secho(
            "1PAY.SYS.DDF01 not available (this is normal on some cards)", fg="yellow"
        )
    except Exception as e:
        click.secho("Error reading 1PAY.SYS.DDF01 (may be normal): " + str(e), fg="red")

    click.secho("\n2PAY.SYS.DDF01 (Index of apps for contactless payments)", bold=True)
    try:
        render_app(card, "2PAY.SYS.DDF01", redact)
    except MissingAppException:
        click.secho(
            "2PAY.SYS.DDF01 not available (this is normal on some cards)", fg="yellow"
        )
    except Exception as e:
        click.secho("Error reading 2PAY.SYS.DDF01 (may be normal): " + str(e), fg="red")

    for app in apps:
        click.secho(
            "\nApplication %s, DF Name: %s"
            % (
                render_element(Tag.APP_LABEL, app[Tag.APP_LABEL]),
                render_element(Tag.DF, app[Tag.ADF_NAME]),
            ),
            bold=True,
        )
        render_app(card, app[Tag.ADF_NAME], redact)

    click.echo("\nFetching card metadata...")
    try:
        tab = SingleTable(card.get_metadata().items())
        tab.inner_heading_row_border = False
        click.echo(tab.table)
    except ErrorResponse as e:
        click.secho("Unable to fetch card data: %s" % e, fg="yellow")


@cli.command(
    help="""[!] Perform EMV CAP authentication.
This will initiate a transaction on the card."""
)
@click.option(
    "--challenge", "-c", metavar="CHALLENGE", help="account number or challenge"
)
@click.option("--amount", "-a", metavar="AMOUNT", help="amount")
@click.pass_context
def cap(ctx, challenge, amount):
    pin = ctx.obj.get("pin", None)
    if not pin:
        click.secho("PIN is required", fg="red")
        sys.exit(2)

    if amount is not None and challenge is None:
        click.secho("Challenge (account number) must be supplied with amount", fg="red")
        sys.exit(3)

    card = get_reader(ctx.obj["reader"])
    try:
        click.echo(card.generate_cap_value(pin, challenge=challenge, value=amount))
    except InvalidPINException:
        click.secho("Invalid PIN", fg="red")
        sys.exit(1)
    except CAPError as e:
        click.secho("Error in CAP generation: %s" % e, fg="red")
        sys.exit(2)


@cli.command(help="List named applications on the card.")
@click.pass_context
def listapps(ctx):
    card = get_reader(ctx.obj["reader"])
    apps = card.list_applications()
    res = [["Index", "Label", "ADF"]]
    i = 0
    for app in apps:
        res.append(
            [
                i,
                render_element(Tag.APP_LABEL, app[Tag.APP_LABEL]),
                render_element(Tag.ADF_NAME, app[Tag.ADF_NAME]),
            ]
        )
        i += 1

    table = SingleTable(res)
    table.title = "Applications"
    click.echo(table.table)


@cli.command(
    help="""[!] Get card processing options and app data.
This will initiate a transaction on the card."""
)
@click.argument("app_index", type=int)
@click.pass_context
def appdata(ctx, app_index):
    redact = ctx.obj["redact"]
    card = get_reader(ctx.obj["reader"])
    apps = card.list_applications()
    app = apps[app_index]
    card.select_application(app[Tag.ADF_NAME])
    click.secho(
        "Selected application %s (%s)"
        % (
            render_element(Tag.APP_LABEL, app[Tag.APP_LABEL]),
            render_element(Tag.ADF_NAME, app[Tag.ADF_NAME]),
        ),
        bold=True,
    )
    opts = card.get_processing_options()

    res = [["Key", "Value"]]
    for k, v in opts.items():
        res.append((k, v))
    table = SingleTable(res)
    table.title = "Processing Options"
    click.echo(table.table)

    app_data = card.get_application_data(opts["AFL"])
    click.echo(as_table(app_data, title="Application Data", redact=redact))



""" EMV Chip Authentication Program, a.k.a DPA, a.k.a Pinsentry.

    There is no public specification for the EMV CAP "standard". The code in this
    module is based on a number of other hacky projects. It works for most UK cards.

    I make no guarantees for non-UK cards as I'm aware that certain banks have
    made their own "customisations" to EMV CAP.
"""
from .protocol.data import Tag
from .protocol.command import GenerateApplicationCryptogramCommand
from .protocol.structures import DOL
from .exc import CAPError
from .util import hex_int

# Older cards will respond with an opaque, packed response to the
# application cryptogram request. This DOL lets us deserialise it.
GAC_RESPONSE_DOL = DOL(
    [
        (Tag((0x9F, 0x27)), 1),  # Cryptogram Information Data
        (Tag((0x9F, 0x36)), 2),  # Application Transaction Counter
        (Tag((0x9F, 0x26)), 8),  # Application Cryptogram
        (Tag((0x9F, 0x10)), 7),  # Issuer Application Data
        (Tag(0x90), 0),
    ]
)

# A static Issuer Proprietary Bitmap which is apparently the default for VISA cards.
# Retrieved from the EMVCAP code:
# https://github.com/zoobab/EMVCAP/blob/0e4877c30972475249e6a2b0253068bfda9e5cf3/EMV-CAP#L519
VISA_STATIC_IPB = [
    0x00,
    0x00,
    0xFF,
    0xFF,
    0xFF,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x20,
    0xB9,
    0x38,
]


def get_arqc_req(app_data, value=None, challenge=None):
    """Generate the data to send with the generate application cryptogram request.
    This data is in the format requested by the card in the CDOL1 field of the
    application data.

    This is the algorithm that barclays_pinsentry.c uses.
    """
    if Tag.CDOL1 not in app_data:
        raise CAPError("Application data doesn't include CDOL1 field: %r" % app_data)

    cdol1 = app_data[Tag.CDOL1]
    data = {
        Tag(0x9A): [0x01, 0x01, 0x01],  # Transaction Date
        Tag(0x95): [0x80, 0x00, 0x00, 0x00, 0x00],  # Terminal Verification Results
    }

    if challenge is not None:
        # If an account number (or challenge) is provided, it goes in the
        # "unpredictable number" field.
        data[Tag((0x9F, 0x37))] = hex_int(challenge)

    if value is not None:
        # If a monetary value is provided, it goes in the "Amount, Authorised"
        # field.
        data[Tag((0x9F, 0x02))] = hex_int(int(round(float(value) * 100, 0)))

    return GenerateApplicationCryptogramCommand(
        GenerateApplicationCryptogramCommand.ARQC, cdol1.serialise(data)
    )


def get_cap_value(response, ipb, psn):
    """Generate a CAP value from the ARQC response.

    The ARQC response is traditionally a data structure returned to the terminal by the
    card during a payment transaction. CAP (mis)uses it to generate an 8-digit one-time
    code. This is done by bitwise masking the values in this structure with the Issuer
    Proprietary Bitmap (IPB) provided by the card in the Application Data structure.
    """

    if Tag.RMTF1 in response.data:
        # Response type 1, deserialise it with our static DOL.
        data = GAC_RESPONSE_DOL.unserialise(response.data[Tag.RMTF1])
    elif Tag.RMTF2 in response.data:
        # Response type 2, TLV format.
        data = response.data[Tag.RMTF2]
    else:
        raise CAPError("Unknown response type in ARQC response: %s" % response.data)

    # IPB acts as a binary mask for the response, where the '0' positions are
    # ignored, and the remaining '1' positions are concatenated, for example:
    # Response: 00101101110100101010010101010101001
    # IPB:      00000111100000000111111110000000000
    # Result:        1011        00101010
    # Concat'd: 101100101010
    # Decimal:  2858

    # Get response data into single list, in the same format as IPB
    resp_data = [item for sublist in data.values() for item in sublist]

    # If the PAN Sequence Number is set, then prepend it to the response data
    if psn is not None:
        resp_data = psn + resp_data

    # Initialise empty string to hold binary result of masking process
    binary_string = ""

    # Iterate through the items in the IPB mask (in reverse)
    for i in reversed(range(0, min(len(ipb), len(resp_data)))):
        # Get the values of data and mask for the current iteration
        data_number = resp_data[i]
        ipb_number = ipb[i]
        # If the mask has some 1s in it
        while ipb_number > 0:
            # Check if the last digit in the binary representation of the mask is '1'
            if ipb_number & 1:
                # And if it is, prepend the last binary digit of the data to the binary string
                binary_string = str(data_number & 1) + binary_string
            # Bitshift both the data and the mask to the right by one bit
            ipb_number >>= 1
            data_number >>= 1

    # And return the binary string converted to a decimal number
    return int(binary_string, 2)



from ..util import format_bytes
import codecs


def assert_valid_byte(val):
    assert type(val) == int
    assert val < 0xFF
    assert val >= 0x00


class CAPDU(object):
    """Command APDU.

    Defined in: EMV 4.3 Book 1 sections:
        - 9.4.1
        - 11.1
    """

    # Map the class name of the command to the CLA,INS bytes.
    # This is derived from EMV 4.3 Book 3 section 6.3.2.
    COMMANDS = {
        "SelectCommand": [0x00, 0xA4],
        "VerifyCommand": [0x00, 0x20],
        "ReadCommand": [0x00, 0xB2],
        "GetDataCommand": [0x80, 0xCA],
        "GenerateApplicationCryptogramCommand": [0x80, 0xAE],
        "GetProcessingOptions": [0x80, 0xA8],
    }

    def marshal(self):
        cla, ins = self.COMMANDS.get(self.__class__.__name__)

        for val in [cla, ins, self.p1, self.p2]:
            assert_valid_byte(val)

        # Mandatory header:
        cmd = [cla, ins, self.p1, self.p2]

        # Conditional body:
        if self.data is not None:
            cmd += [len(self.data)]  # Lc
            cmd += self.data

        # Bytes expected:
        if self.le is not None:
            cmd += [self.le]  # Le
        return cmd

    @classmethod
    def get_class(cls, pdu_bytes):
        for clsname, b in cls.COMMANDS.items():
            if pdu_bytes == b:
                return globals()[clsname]
        return None

    @classmethod
    def unmarshal(cls, data):
        pcls = cls.get_class(data[:2])
        obj = object.__new__(pcls)
        obj.p1 = data[2]
        obj.p2 = data[3]
        if len(data) > 5:
            obj.lc = data[4]
            obj.data = data[5 : obj.lc + 5]
        if len(data) > obj.lc + 5:
            obj.le = data[-1]
        else:
            obj.le = None
        return obj

    def __repr__(self):
        return "<Command[%s] P1: %02x, P2: %02x, data: %s, Le: %02x>" % (
            self.name,
            self.p1,
            self.p2,
            format_bytes(self.data),
            self.le or 0,
        )


class SelectCommand(CAPDU):
    """Select an application or file on the card.

    Defined in: EMV 4.3 Book 1 section 11.3
    """

    name = "Select"

    def __init__(self, file_path=None, file_identifier=None, next_occurrence=False):
        if file_path is not None:
            if isinstance(file_path, str):
                self.data = [ord(c) for c in file_path]
            else:
                self.data = file_path
            self.p1 = 0x04  # Select by path
        else:
            self.data = file_identifier
            self.p1 = 0x00

        if next_occurrence:
            self.p2 = 0x02  # Next occurrence
        else:
            self.p2 = 0x00  # First or only occurrence

        self.le = 0x00

    def __repr__(self):
        data = " ".join(["%02x" % i for i in self.data])
        return "<Command[%s] P1: %02x, P2: %02x, File name: %s, Le: %02x>" % (
            self.name,
            self.p1,
            self.p2,
            data.upper(),
            self.le or 0,
        )


class ReadCommand(CAPDU):
    """Read a record from an application or file.

    Defined in: EMV 4.3 Book 1 section 11.2
    """

    name = "Read"

    P2_RECORD_NUMBER = 0x04  # P1 is a record number

    def __init__(self, record_number, sfi=None):
        assert type(sfi) == int or sfi is None
        assert type(record_number) == int

        self.p1 = record_number
        if sfi is not None:
            self.p2 = (sfi << 3) + self.P2_RECORD_NUMBER
        else:
            self.p2 = 0x04

        self.data = None
        self.le = 0x00


class GetDataCommand(CAPDU):
    """Get miscellaneous data

    Defined in: EMV 4.3 Book 3 section 6.5.7
    """

    name = "Get Data"

    ATC = (0x9F, 0x36)
    LAST_ONLINE_ATC = (0x9F, 0x13)
    PIN_TRY_COUNT = (0x9F, 0x17)
    LOG_FORMAT = (0x9F, 0x4F)

    def __init__(self, obj):
        assert type(obj) == tuple
        self.p1 = obj[0]
        self.p2 = obj[1]
        self.data = None
        self.le = 0x00


class VerifyCommand(CAPDU):
    """Verify the PIN.

    Defined in: EMV 4.3 Book 3 section 6.5.12
    """

    name = "Verify"

    PIN_PLAINTEXT = 0b10000000
    PIN_ENCIPHERED = 0b10001000

    def __init__(self, pin):
        assert 4 <= len(str(pin)) <= 12

        self.p1 = 0x00
        self.p2 = self.PIN_PLAINTEXT

        # Fields in this data structure are split on a nibble boundary,
        # so assemble as a hex string and decode
        data = ("2%x%s" % (len(str(pin)), pin)).encode("ascii")
        while len(data) < 16:
            data += b"f"
        self.data = codecs.decode(data, "hex")

        self.le = 0x00


class GenerateApplicationCryptogramCommand(CAPDU):
    """Defined in: EMV 4.3 Book 3 section 6.5.5"""

    name = "Generate Application Cryptogram"

    # Table 12
    AAC = 0b00000000
    TC = 0b01000000
    ARQC = 0b10000000
    CDA_SIG = 0b00010000

    def __init__(self, crypto_type, data, cda_sig=False):

        self.p1 = crypto_type
        if cda_sig:
            self.p1 |= self.CDA_SIG

        self.data = data

        self.p2 = 0x00
        self.le = 0x00


class GetProcessingOptions(CAPDU):
    """Defined in: EMV 4.3 Book 3 section 6.5.8"""

    name = "Get Processing Opts"

    def __init__(self, pdol=None):
        self.p1 = 0x00
        self.p2 = 0x00
        if pdol is None:
            self.data = [0x83, 0x00]
        else:
            self.data = pdol
        self.le = 0x00



from .structures import TLV


class RAPDU(object):
    """Response Application Protocol Data Unit

    Defined in: EMV 4.3 Book 3 section 6.3.3
    """

    @classmethod
    def unmarshal(cls, data):
        assert len(data) > 1
        sw1 = data[-2]
        sw2 = data[-1]

        assert sw1 not in (0x61, 0x6C)  # should be handled by the transport layer.

        if sw1 == 0x90 and sw2 == 0x00:
            obj = SuccessResponse()
        elif sw1 in (0x62, 0x63):
            obj = WarningResponse()
        else:
            obj = ErrorResponse()
        obj.sw1 = sw1
        obj.sw2 = sw2
        if len(data) > 2:
            obj.data = TLV.unmarshal(data[:-2])
        else:
            obj.data = None

        if type(obj) == ErrorResponse:
            raise obj

        return obj

    def get_status(self):
        return "SW1: %02x, SW2: %02x" % (self.sw1, self.sw2)

    def __repr__(self):
        res = '<%s: "%s"' % (self.__class__.__name__, self.get_status())
        if self.data:
            res += ", data: " + str(self.data)
        return res + ">"

    def __str__(self):
        return repr(self)


class SuccessResponse(RAPDU):
    def get_status(self):
        return "Process completed"


class WarningResponse(RAPDU):
    def get_status(self):
        if self.sw1 == 0x62 and self.sw2 == 0x83:
            return "State of non-volatile memory unchanged; selected file invalidated"
        elif self.sw1 == 0x63 and self.sw2 == 0x00:
            return "State of non-volatile memory changed; authentication failed"
        elif self.sw1 == 0x63 and self.sw2 & 0xC0 == 0xC0:
            return "State of non-volatile memory changed; counter is %i" % (
                self.sw2 & ~0xC0
            )


class ErrorResponse(RAPDU, Exception):
    ERRORS = {
        (0x69, 0x83): "Command not allowed; authentication method blocked",
        (0x69, 0x84): "Command not allowed; referenced data invalidated",
        (0x69, 0x85): "Command not allowed; conditions of use not satisfied",
        (0x6A, 0x81): "Wrong parameter(s) P1 P2; function not supported",
        (0x6A, 0x82): "Wrong parameter(s) P1 P2; file not found",
        (0x6A, 0x83): "Wrong parameter(s) P1 P2; record not found",
        (0x6A, 0x86): "Inconsistent parameters P1-P2",
        (0x6A, 0x87): "Lc inconsistent with P1-P2",
        (0x6A, 0x88): "Referenced data (data objects) not found",
        (0x6D, 0x00): "Instruction code not supported or invalid",
    }

    def get_status(self):
        if (self.sw1, self.sw2) in self.ERRORS:
            return self.ERRORS.get((self.sw1, self.sw2))
        else:
            return "Unknown error: %02x %02x" % (self.sw1, self.sw2)



import re


def hex_int(val):
    """Convert an integer into a decimal-encoded hex integer as bytes,
    which the EMV spec seems awfully keen on.

    >>> hex_int(123456)
    [0x12, 0x34, 0x56]
    >>> hex_int(65432)
    [0x06, 0x54, 0x32]
    """
    s = str(val)
    if len(s) % 2 != 0:
        s = "0" + s
    return [int(s[i : i + 2], 16) for i in range(0, len(s), 2)]


def from_hex_int(val):
    """Convert hex digits to decimal.

    >>> from_hex_int([0x12, 0x34])
    1234
    """
    return int("".join(["%02x" % i for i in val]))


def from_hex_date(val):
    return "%02x/%02x/%02x" % (val[0], val[1], val[2])


def decode_int(val):
    result = val[0]
    for i in val[1:]:
        result = result << 8
        result += i
    return result


def format_bytes(data):
    if type(data) == int:
        return "[%02X]" % data
    return "[" + " ".join(["%02X" % i for i in data]) + "]"


def unformat_bytes(data):
    data = re.split(r"(?:\s+|:)", data)
    return [int(i, 16) for i in data]


def bit_set(value, bit):
    mask = 1 << bit
    return (value & mask) == mask



import logging
from .transmission import TransmissionProtocol
from .protocol.structures import TLV
from .protocol.data import Tag
from .protocol.response import WarningResponse, ErrorResponse
from .protocol.command import (
    SelectCommand,
    ReadCommand,
    GetDataCommand,
    GetProcessingOptions,
    VerifyCommand,
)
from .exc import InvalidPINException, MissingAppException
from .util import decode_int
from .cap import get_arqc_req, get_cap_value, VISA_STATIC_IPB

log = logging.getLogger(__name__)


class Card(object):
    """High-level card manipulation API"""

    def __init__(self, connection):
        self.tp = TransmissionProtocol(connection)

    def get_mf(self):
        """Get the master file (MF)."""
        return self.tp.exchange(SelectCommand(file_identifier=[0x3F, 0x00]))

    def get_pse(self):
        """Get the Payment System Environment (PSE) file"""
        return self.tp.exchange(SelectCommand("1PAY.SYS.DDF01"))

    def list_applications(self):
        """List applications on the card"""
        try:
            return self._list_applications_sfi()
        except ErrorResponse:
            return self._list_applications_static_aid()

    def _list_applications_static_aid(self):
        """Try to find applications by trying to select a static application ID.
        This is an older method of app discovery which is still used by some cards.
        """
        STATIC_AIDS = [
            [0xA0, 0x00, 0x00, 0x00, 0x25, 0x01],  # Amex
            [0xA0, 0x00, 0x00, 0x00, 0x03, 0x10, 0x10],  # Visa
            [0xA0, 0x00, 0x00, 0x00, 0x04, 0x10, 0x10],  # Mastercard
        ]
        apps = []
        for aid in STATIC_AIDS:
            try:
                res = self.tp.exchange(SelectCommand(aid))

                # This is a bit of a hack, we transform this response into something which looks
                # like the result from the SFI method, so that callers of list_applications get a
                # consistent result.
                apps.append(
                    TLV(
                        {
                            Tag.ADF_NAME: res.data[Tag.FCI][Tag.DF],
                            Tag.APP_LABEL: res.data[Tag.FCI][Tag.FCI_PROP][
                                Tag.APP_LABEL
                            ],
                        }
                    )
                )
            except ErrorResponse:
                continue
        return apps

    def _list_applications_sfi(self):
        """List applications on the card using the SFI method.

        This fetches the SFI (short file identifier) from the PSE (Payment System Environment)
        file, and uses it to locate all the apps on the card.
        """
        pse = self.get_pse()
        sfi = pse.data[Tag.FCI][Tag.FCI_PROP][Tag.SFI][0]
        apps = []

        # Apps may be stored in different records, so iterate through records
        # until we hit an error
        for i in range(1, 31):
            try:
                res = self.read_record(i, sfi)
            except ErrorResponse:
                break
            new_apps = res.data[Tag.RECORD][Tag.APP]
            if type(new_apps) is not list:
                new_apps = [new_apps]
            apps += new_apps
        return apps

    def read_record(self, record_number, sfi=None):
        return self.tp.exchange(ReadCommand(record_number, sfi))

    def select_application(self, app):
        try:
            res = self.tp.exchange(SelectCommand(app))
        except ErrorResponse as e:
            raise MissingAppException(e)
        return res

    def get_data_item(self, item, tag):
        try:
            res = self.tp.exchange(GetDataCommand(item))
            return res.data[tag]
        except ErrorResponse:
            return None

    def get_metadata(self):
        data = {}
        res = self.get_data_item(GetDataCommand.PIN_TRY_COUNT, (0x9F, 0x17))
        if res:
            data["pin_retries"] = res[0]

        res = self.get_data_item(GetDataCommand.ATC, (0x9F, 0x36))
        if res:
            data["atc"] = decode_int(res)

        res = self.get_data_item(GetDataCommand.LAST_ONLINE_ATC, (0x9F, 0x13))
        if res:
            data["last_online_atc"] = decode_int(res)

        return data

    def get_processing_options(self):
        res = self.tp.exchange(GetProcessingOptions())
        if Tag.RMTF1 in res.data:
            # Response template format 1
            return {"AIP": res.data[Tag.RMTF1][:2], "AFL": res.data[Tag.RMTF1][2:]}
        elif Tag.RMTF2 in res.data:
            # Response template format 2
            return {"AIP": res.data[Tag.RMTF2][0x82], "AFL": res.data[Tag.RMTF2][0x94]}

    def get_application_data(self, afl):
        assert len(afl) % 4 == 0
        data = TLV()
        for i in range(0, len(afl), 4):
            sfi = afl[i] >> 3
            start_rec = afl[i + 1]
            end_rec = afl[i + 2]
            # dar = afl[i + 3]
            for i in range(start_rec, end_rec + 1):
                res = self.tp.exchange(ReadCommand(i, sfi))
                data.update(res.data[Tag.RECORD])
        return data

    def verify_pin(self, pin):
        """Verify the PIN, raising an exception if it fails."""
        res = self.tp.exchange(VerifyCommand(pin))
        if type(res) == WarningResponse:
            raise InvalidPINException(str(res))

        return res

    def generate_cap_value(self, pin, challenge=None, value=None):
        """Perform a transaction to generate the EMV CAP (Pinsentry) value."""
        apps = self.list_applications()

        if len(apps) == 0:
            raise MissingAppException("No apps on card")

        # We're selecting the last app on the card here, which seems be the correct
        # (bank-specific) one. If this isn't always the case, it may be better to
        # select the app with ADF [A0 00 00 00 03 80 02].
        self.select_application(apps[-1][Tag.ADF_NAME])

        # Get Processing Options starts the transaction on the card and
        # increments the transaction counter.
        opts = self.get_processing_options()

        # Fetch the application data referenced in the processing options.
        # This includes the CDOL data structure which dictates the format
        # of the data passed to the Get Application Cryptogram function.
        app_data = self.get_application_data(opts["AFL"])

        self.verify_pin(pin)

        resp = self.tp.exchange(
            get_arqc_req(app_data, challenge=challenge, value=value)
        )

        # Set default: don't use PAN Sequence Number
        psn = None

        # If the third bit of Issuer Authentication Flags is set then use the PAN Sequence Number
        if Tag.IAF in app_data and app_data[Tag.IAF][0] & 0x40:
            psn = app_data[Tag.PAN_SN]

        # Fetch the Issuer Proprietary Bitmap. In most UK cards this is provided in the application data file.
        #
        # In some cases the IPB may not be present. This static IPB is from the EMVCAP code.
        #
        # It appears that Belgian cards use their own silliness.
        # https://github.com/zoobab/EMVCAP/blob/master/EMV-CAP#L512
        if Tag.IPB in app_data:
            ipb = app_data[Tag.IPB]
        else:
            log.warn(
                "Issuer Proprietary Bitmap not found on card - using static VISA IPB. "
                "The resulting code may not work!"
            )
            ipb = VISA_STATIC_IPB

        return get_cap_value(resp, ipb, psn)



from collections import OrderedDict
from .data import (
    ELEMENT_FORMAT,
    render_element,
    read_tag,
    read_length,
    is_constructed,
    Tag,
)
from .data_elements import Parse, EPC_PRODUCT_ID
from ..util import decode_int, bit_set
import logging

log = logging.getLogger(__name__)


def parse_element(tag, value):
    if ELEMENT_FORMAT.get(tag) == Parse.DOL:
        value = DOL.unmarshal(value)
    elif ELEMENT_FORMAT.get(tag) == Parse.TAG_LIST:
        value = TagList.unmarshal(value)
    elif ELEMENT_FORMAT.get(tag) == Parse.ASRPD:
        value = ASRPD.unmarshal(value)
    elif ELEMENT_FORMAT.get(tag) == Parse.CVM_LIST:
        value = CVMList.unmarshal(value)
    elif ELEMENT_FORMAT.get(tag) == Parse.AUC:
        value = AUC.unmarshal(value)
    return value


class TLV(OrderedDict):
    """BER-TLV
    A serialisation format.

    Documented in EMV 4.3 Book 3 Annex B
    """

    @classmethod
    def unmarshal(cls, data):
        tlv = cls()
        i = 0

        if len(data) < 3:
            # A valid TLV record is at least three bytes, anything less is probably a bug.
            # I've seen some cards present this (with a TLV of simply [0x61]), so silently ignore.
            log.info("Invalid TLV - too short: %s", data)
            return data

        while i < len(data):
            tag, tag_len = read_tag(data[i:])
            i += tag_len
            if len(data) <= i:
                log.info("Invalid TLV - read beyond end of buffer at %s: %s", tag, data)
                return tlv

            length, length_len = read_length(data[i:])
            i += length_len

            value = data[i : i + length]

            if is_constructed(tag[0]):
                value = TLV.unmarshal(value)

            tag = Tag(tag)
            value = parse_element(tag, value)

            # If we have duplicate tags, make them into a list
            if tag in tlv:
                if type(tlv[tag]) is not list:
                    tlv[tag] = [tlv[tag]]
                tlv[tag].append(value)
            else:
                tlv[tag] = value
            i += length
        return tlv

    def __repr__(self):
        vals = []
        for key, val in self.items():
            out = "\n%s: " % str(key)
            if (
                type(val) in (TLV, DOL)
                or type(val) is list
                and len(val) > 0
                and type(val[0]) in (TLV, DOL)
            ):
                out += str(val)
            else:
                out += render_element(key, val)
            vals.append(out)
        return "{" + (", ".join(vals)) + "}"


class ASRPD(dict):
    """Application Selection Registered Proprietary Data list.

    An almost-TLV structure used in the FCI Discretionary Data object.
    The tags here are fixed-length.

    https://www.emvco.com/wp-content/uploads/2017/05/BookB_Entry_Point_Specification_v2_6_20160809023257319.pdf
    """

    @classmethod
    def unmarshal(cls, data):
        asrpd = cls()

        i = 0

        while i < len(data):
            # 2 bytes Proprietary Data Identifier
            pdi = "%02i%02i" % tuple(data[i : i + 2])
            i += 2

            length = data[i]
            i += 1

            asrpd[pdi] = data[i : i + length]
            i += length

        return asrpd

    def __repr__(self):
        ret = "<ASRPD: "
        for pdi, value in self.items():
            if pdi == "0001":
                ret += "Electronic Product Identification: "
                ret += EPC_PRODUCT_ID.get(value[0], "Unknown")
        ret += ">"
        return ret


class DOL(list):
    """Data Object List.
    This is sent by the card to the terminal to define a structure for
    future transactions, consisting of an ordered list of data elements and lengths.

    It's essentially a TLV object without the values.

    EMV 4.3 Book 3 section 5.4"""

    @classmethod
    def unmarshal(cls, data):
        """Construct a DOL object from the binary representation (as a list of bytes)"""
        dol = cls()
        i = 0
        while i < len(data):
            tag, tag_len = read_tag(data[i:])
            i += tag_len
            length = data[i]
            i += 1
            dol.append((Tag(tag), length))
        return dol

    def size(self):
        """Total size of the resulting structure in bytes."""
        return sum([val[1] for val in self])

    def unserialise(self, data):
        """Parse an input stream of bytes and return a TLV object."""
        if self.size() != len(data):
            raise Exception(
                "Incorrect input size (expecting %s bytes, got %s)"
                % (self.size(), len(data))
            )

        tlv = TLV()
        i = 0
        for tag, length in self:
            tlv[tag] = data[i : i + length]
            i += length

        return tlv

    def __contains__(self, val):
        for v in self:
            if v[0] == val:
                return True
        return False

    def serialise(self, data):
        """Given a dictionary of tag -> value, write this data out
        according to the DOL. Missing data will be null.
        """
        output = []
        for tag, length in self:
            value = data.get(tag, [0x0] * length)
            if len(value) < length:
                # If the length is shorter than required, left-pad it.
                value = [0x0] * (length - len(value)) + value
            elif len(value) > length:
                raise Exception("Data for tag %s is too long" % tag)
            output += value

        assert len(output) == self.size()
        return output


class TagList(list):
    """A list of tags."""

    @classmethod
    def unmarshal(cls, data):
        tag_list = cls()
        i = 0
        while i < len(data):
            tag, tag_len = read_tag(data[i:])
            i += tag_len
            tag_list.append(Tag(tag))
        return tag_list


class CVMRule(object):
    """EMV 4.3 book 3 appendix C3"""

    RULES = {
        # 0b00000000: "Fail CVM processing",
        0b00000001: "Plaintext PIN verification performed by ICC",
        0b00000010: "Enciphered PIN verified online",
        0b00000011: "Plaintext PIN verification performed by ICC and signature (paper)",
        0b00000100: "Enciphered PIN verification performed by ICC",
        0b00000101: "Enciphered PIN verification performed by ICC and signature (paper)",
        0b00011110: "Signature (paper)",
        0b00111111: "No CVM required",
    }

    CODES = {
        0: "Always",
        1: "If unattended cash",
        2: "If not unattended cash and not manual cash and not purchase with cashback",
        3: "If terminal supports the CVM",
        4: "If manual cash",
        5: "If purchase with cashback",
        6: "If transaction is in the application currency and is under X value",
        7: "If transaction is in the application currency and is over X value",
        8: "If transaction is in the application currency and is under Y value",
        9: "If transaction is in the application currency and is over Y value",
    }

    @classmethod
    def unmarshal(cls, b1, b2):
        rule = cls()
        rule.b1 = b1
        rule.b2 = b2
        return rule

    def rule_repr(self):
        for k, v in self.RULES.items():
            if (self.b1 & k) == k:
                return v
        return "Fail CVM processing"

    def code_repr(self):
        return self.CODES.get(self.b2)

    def fail_if_unsuccessful(self):
        return self.b1 & 0b01000000 == 0b01000000

    def __repr__(self):
        if self.fail_if_unsuccessful():
            fail = ". Else, fail verification."
        else:
            fail = ""

        return "%s, %s%s" % (self.code_repr(), self.rule_repr(), fail)


class CVMList(object):
    """CVM is a tiny language for the card to dictate when the terminal should fail the
        transaction or force it online.

        It doesn't seem to get much interesting use on many of the cards I've seen though.

    EMV 4.3 book 3 section 10.5"""

    def __init__(self):
        self.x = None
        self.y = None
        self.rules = []

    @classmethod
    def unmarshal(cls, data):
        cvm_list = cls()
        if len(data) < 10 or len(data) % 2 != 0:
            return cvm_list

        cvm_list.x = decode_int(data[0:4])
        cvm_list.y = decode_int(data[4:8])

        cvm_list.rules = []

        i = 8
        while i < len(data):
            cvm_list.rules.append(CVMRule.unmarshal(data[i], data[i + 1]))
            i += 2

        return cvm_list

    def __repr__(self):
        return "<CVM List x: %s, y: %s, rules: %s>" % (
            self.x,
            self.y,
            "; ".join([repr(r) for r in self.rules]),
        )


class AUC(object):

    B1_FIELDS = [
        "Valid for domestic cash transactions",
        "Valid for international cash transactions",
        "Valid for domestic goods",
        "Valid for international goods",
        "Valid for domestic services",
        "Valid for international services",
        "Valid at ATMs",
        "Valid at terminals other than ATMs",
    ]

    B2_FIELDS = ["Domestic cashback allowed", "International cashback allowed"]

    @classmethod
    def unmarshal(cls, data):
        auc = cls()
        if len(data) != 2:
            return auc

        auc.b1 = data[0]
        auc.b2 = data[1]
        return auc

    def get_uses(self):
        uses = []
        for i in range(0, len(self.B1_FIELDS)):
            if bit_set(self.b1, i):
                uses.append(self.B1_FIELDS[i])

        for i in range(0, len(self.B2_FIELDS)):
            if bit_set(self.b2, i):
                uses.append(self.B2_FIELDS[i])
        return uses

    def __repr__(self):
        return "<AUC: %s>" % ", ".join(self.get_uses())



import logging
from .protocol.response import RAPDU
from .util import format_bytes


class TransmissionProtocol(object):
    """Transport layer. Only currently supports T0 transport.

    Defined in: EMV 4.3 Book 1 section 9
    See also Annex A for examples.
    """

    def __init__(self, connection):
        """Connection should be a pyscard connection."""
        self.log = logging.getLogger(__name__)
        self.connection = connection
        self.connection.connect(connection.T0_protocol)
        assert connection.getProtocol() == connection.T0_protocol
        self.log.info("Connected to reader")

    def transmit(self, tx_data):
        """Send raw data to the card, and receive the reply.

        tx_data should be a list of bytes.

        Returns a tuple of (data, sw1, sw2) where sw1 and sw2
        are the protocol status bytes.
        """
        self.log.debug("Tx: %s", format_bytes(tx_data))
        data, sw1, sw2 = self.connection.transmit(tx_data)
        self.log.debug("Rx: %s, SW1: %02x, SW2: %02x", format_bytes(data), sw1, sw2)
        return data, sw1, sw2

    def exchange(self, capdu):
        """Send a command to the card and return the response.

        Accepts a CAPDU object and returns a RAPDU.
        """
        send_data = capdu.marshal()
        data, sw1, sw2 = self.transmit(send_data)

        if sw1 == 0x6C:
            # ICC asks to reduce data size requested
            send_data[4] = sw2
            data, sw1, sw2 = self.transmit(send_data)

        while sw1 == 0x61:
            # ICC has continuation data
            d, sw1, sw2 = self.transmit([0x00, 0xC0, 0x00, 0x00, sw2])
            data = data[:-2] + d

        res = RAPDU.unmarshal(data + [sw1, sw2])
        return res



from enum import Enum


class Parse(Enum):
    BYTES = 1
    ASCII = 2
    DOL = 3  # Data Object List
    DEC = 4  # Decimal encoded in hex
    DATE = 5  # Date encoded in hex digits as [0xYY 0xMM 0xDD]
    INT = 6  # Decode bytes as an integer
    COUNTRY = 7  # Decode bytes as an ISO country code
    CURRENCY = 8  # ISO currency code
    TAG_LIST = 9  # A list of tag names
    ASRPD = 10  # Application Selection Registered Proprietary Data list
    CVM_LIST = 11  # Cardholder Verification Method List
    AUC = 12  # Application Usage Control


# This table is derived from:
#    - EMV 4.3 Book 3 Annex A
#    - EMV 4.1 Book 1 Annex B
#
#   (Tag, Description, Parsing, Shortname)
ELEMENT_TABLE = [
    (0x42, "Issuer Identification Number", Parse.BYTES, "IIN"),
    (0x46, "Pre-issuing data", Parse.BYTES, None),
    (0x4F, "Application Dedicated File (ADF) Name", Parse.BYTES, "ADF_NAME"),
    (0x50, "Application Label", Parse.ASCII, "APP_LABEL"),
    (0x57, "Track 2 Equivalent Data", Parse.BYTES, "TRACK2"),
    (0x5A, "Application Primary Account Number (PAN)", Parse.DEC, "PAN"),
    (0x5E, "Proprietary Login Data", Parse.BYTES, None),
    ((0x5F, 0x20), "Cardholder Name", Parse.ASCII, None),
    ((0x5F, 0x24), "Application Expiration Date", Parse.DATE, None),
    ((0x5F, 0x25), "Application Effective Date", Parse.DATE, None),
    ((0x5F, 0x28), "Issuer Country Code", Parse.COUNTRY, None),
    ((0x5F, 0x2A), "Transaction Country Code", Parse.COUNTRY, None),
    ((0x5F, 0x2D), "Language Preference", Parse.ASCII, None),
    ((0x5F, 0x30), "Service Code", Parse.BYTES, None),
    (
        (0x5F, 0x34),
        "Application Primary Account Number (PAN) Sequence Number",
        Parse.INT,
        "PAN_SN",
    ),
    ((0x5F, 0x36), "Transaction Currency Exponent", Parse.BYTES, None),
    ((0x5F, 0x50), "Issuer URL", Parse.ASCII, None),
    ((0x5F, 0x53), "International Bank Account Number (IBAN)", Parse.BYTES, "IBAN"),
    ((0x5F, 0x54), "Bank Identifier Code (BIC)", Parse.BYTES, "BIC"),
    ((0x5F, 0x55), "Issuer Country Code (alpha2 format)", Parse.ASCII, None),
    ((0x5F, 0x56), "Issuer Country Code (alpha3 format)", Parse.ASCII, None),
    ((0x5F, 0x57), "Account Type", Parse.BYTES, None),
    (0x61, "Application Template", Parse.BYTES, "APP"),
    (0x6F, "FCI Template", Parse.BYTES, "FCI"),
    (0x70, "Read Record Response Template", Parse.BYTES, "RECORD"),
    (0x71, "Issuer Script Template 1", Parse.BYTES, None),
    (0x72, "Issuer Script Template 2", Parse.BYTES, None),
    (0x73, "Directory Discretionary Template", Parse.BYTES, None),
    (0x77, "Response Template Format 2", Parse.BYTES, "RMTF2"),
    (0x80, "Response Template Format 1", Parse.BYTES, "RMTF1"),
    (0x81, "Amount, Authorised (Binary)", Parse.BYTES, None),
    (0x82, "Application Interchange Profile", Parse.BYTES, None),
    (0x83, "Command Template", Parse.BYTES, None),
    (0x84, "Dedicated File (DF) Name", Parse.BYTES, "DF"),
    (0x86, "Issuer Script Command", Parse.BYTES, None),
    (0x87, "Application Priority Indicator", Parse.INT, None),
    (0x88, "Short File Identifier", Parse.BYTES, "SFI"),
    (0x89, "Authorisation Code", Parse.BYTES, None),
    (0x8A, "Authorisation Response Code", Parse.BYTES, None),
    (0x8C, "Card Risk Management Data Object List 1 (CDOL1)", Parse.DOL, "CDOL1"),
    (0x8D, "Card Risk Management Data Object List 2 (CDOL2)", Parse.DOL, "CDOL2"),
    (0x8E, "Cardholder Verification Method (CVM) List", Parse.CVM_LIST, None),
    (0x8F, "Certification Authority Public Key Index", Parse.BYTES, None),
    (0x90, "Issuer Public Key Certificate", Parse.BYTES, None),
    (0x91, "Issuer Authentication Data", Parse.BYTES, None),
    (0x92, "Issuer Public Key Remainder", Parse.BYTES, None),
    (0x93, "Signed Static Application Data", Parse.BYTES, None),
    (0x94, "Application File Locator", Parse.BYTES, "AFL"),
    (0x95, "Terminal Verification Results", Parse.BYTES, None),
    (0x97, "Transaction Certificate Data Object List (TDOL)", Parse.DOL, "TDOL"),
    (0x98, "Transaction Certificate (TC) Hash Value", Parse.BYTES, None),
    (0x99, "Transaction Personal Identification Number (PIN) Data", Parse.BYTES, None),
    (0x9A, "Transaction Date", Parse.DATE, None),
    (0x9B, "Transaction Status Information", Parse.BYTES, None),
    (0x9C, "Transaction Type", Parse.BYTES, None),
    (0x9D, "DDF Name", Parse.BYTES, "DDF"),
    (0xC8, "Card risk management country code", Parse.COUNTRY, None),
    (0xC9, "Card risk management currency code", Parse.CURRENCY, None),
    ((0x9F, 0x01), "Acquirer Identifier", Parse.BYTES, None),
    ((0x9F, 0x02), "Amount, Authorised", Parse.BYTES, None),
    ((0x9F, 0x03), "Amount, Other (Numeric)", Parse.BYTES, None),
    ((0x9F, 0x04), "Amount, Other (Binary)", Parse.BYTES, None),
    ((0x9F, 0x05), "Application Discretionary Data", Parse.BYTES, None),
    ((0x9F, 0x06), "Application Identifier (AID) - terminal", Parse.BYTES, None),
    ((0x9F, 0x07), "Application Usage Control", Parse.AUC, "AUC"),
    ((0x9F, 0x08), "Application Version Number", Parse.BYTES, None),
    ((0x9F, 0x09), "Application Version Number", Parse.BYTES, None),
    (
        (0x9F, 0x0A),
        "Application Selection Registered Proprietary Data",
        Parse.ASRPD,
        None,
    ),
    ((0x9F, 0x0B), "Cardholder Name Extended", Parse.ASCII, None),
    ((0x9F, 0x0D), "Issuer Action Code - Default", Parse.BYTES, None),
    ((0x9F, 0x0E), "Issuer Action Code - Denial", Parse.BYTES, None),
    ((0x9F, 0x0F), "Issuer Action Code - Online", Parse.BYTES, None),
    ((0x9F, 0x10), "Issuer Application Data", Parse.BYTES, None),
    ((0x9F, 0x11), "Issuer Code Table Index", Parse.BYTES, None),
    ((0x9F, 0x12), "Application Preferred Name", Parse.ASCII, None),
    (
        (0x9F, 0x13),
        "Last Online Application Transaction Counter (ATC) Register",
        Parse.INT,
        None,
    ),
    ((0x9F, 0x14), "Lower Consecutive Offline Limit", Parse.BYTES, None),
    ((0x9F, 0x15), "Merchant Category Code", Parse.BYTES, None),
    ((0x9F, 0x16), "Merchant Identifier", Parse.BYTES, None),
    ((0x9F, 0x17), "PIN Try Counter", Parse.INT, None),
    ((0x9F, 0x18), "Issuer Script Identifier", Parse.BYTES, None),
    ((0x9F, 0x1A), "Terminal Country Code", Parse.COUNTRY, None),
    ((0x9F, 0x1B), "Terminal Floor Limit", Parse.BYTES, None),
    ((0x9F, 0x1C), "Terminal Identification", Parse.BYTES, None),
    ((0x9F, 0x1D), "Terminal Risk Management Data", Parse.BYTES, None),
    ((0x9F, 0x1E), "Interface Device (IFD) Serial Number", Parse.BYTES, None),
    ((0x9F, 0x1F), "Track 1 Discretionary Data", Parse.BYTES, None),
    ((0x9F, 0x20), "Track 2 Discretionary Data", Parse.BYTES, None),
    ((0x9F, 0x21), "Transaction Time", Parse.BYTES, None),
    ((0x9F, 0x22), "Certification Authority Public Key Index", Parse.BYTES, None),
    ((0x9F, 0x23), "Upper Consecutive Offline Limit", Parse.BYTES, None),
    ((0x9F, 0x26), "Application Cryptogram", Parse.BYTES, None),
    ((0x9F, 0x27), "Cryptogram Information Data", Parse.BYTES, None),
    ((0x9F, 0x2D), "ICC PIN Encipherment Public Key Certificate", Parse.BYTES, None),
    ((0x9F, 0x2E), "ICC PIN Encipherment Public Key Exponent", Parse.BYTES, None),
    ((0x9F, 0x2F), "ICC PIN Encipherment Public Key Remainder", Parse.BYTES, None),
    ((0x9F, 0x32), "Issuer Public Key Exponent", Parse.BYTES, None),
    ((0x9F, 0x33), "Terminal Capabilities", Parse.BYTES, None),
    ((0x9F, 0x34), "Cardholder Verification Method (CVM) Results", Parse.BYTES, None),
    ((0x9F, 0x35), "Terminal Type", Parse.BYTES, None),
    ((0x9F, 0x36), "Application Transaction Counter", Parse.INT, "ATC"),
    ((0x9F, 0x37), "Unpredictable Number", Parse.BYTES, None),
    ((0x9F, 0x38), "Processing Options Data Object List (PDOL)", Parse.DOL, "PDOL"),
    ((0x9F, 0x39), "Point-of-Service (POS) Entry Mode", Parse.BYTES, None),
    ((0x9F, 0x3A), "Amount, Reference Currency", Parse.BYTES, None),
    ((0x9F, 0x3B), "Application Reference Currency", Parse.BYTES, None),
    ((0x9F, 0x3C), "Transaction Reference Currency Code", Parse.BYTES, None),
    ((0x9F, 0x3D), "Transaction Reference Currency Exponent", Parse.BYTES, None),
    ((0x9F, 0x40), "Additional Terminal Capabilities", Parse.BYTES, None),
    ((0x9F, 0x41), "Transaction Sequence Counter", Parse.BYTES, None),
    ((0x9F, 0x42), "Application Currency Code", Parse.CURRENCY, None),
    ((0x9F, 0x43), "Application Reference Currency Exponent", Parse.BYTES, None),
    ((0x9F, 0x44), "Application Currency Exponent", Parse.INT, None),
    ((0x9F, 0x45), "Data Authentication Code", Parse.BYTES, None),
    ((0x9F, 0x46), "ICC Public Key Certificate", Parse.BYTES, None),
    ((0x9F, 0x47), "ICC Public Key Exponent", Parse.BYTES, None),
    ((0x9F, 0x48), "ICC Public Key Remainder", Parse.BYTES, None),
    (
        (0x9F, 0x49),
        "Dynamic Data Authentication Data Object List (DDOL)",
        Parse.DOL,
        "DDOL",
    ),
    ((0x9F, 0x4A), "Static Data Authentication Tag List", Parse.TAG_LIST, None),
    ((0x9F, 0x4B), "Signed Dynamic Application Data", Parse.BYTES, None),
    ((0x9F, 0x4C), "ICC Dynamic Number", Parse.BYTES, None),
    ((0x9F, 0x4D), "Log Entry", Parse.BYTES, None),
    ((0x9F, 0x4E), "Merchant Name and Location", Parse.BYTES, None),
    ((0x9F, 0x4F), "Log Format", Parse.BYTES, None),
    ((0x9F, 0x55), "Issuer Authentication Flags", Parse.BYTES, "IAF"),
    ((0x9F, 0x56), "Issuer Proprietary Bitmap", Parse.BYTES, "IPB"),
    (
        (0x9F, 0x5C),
        "Cumulative Total Transaction Amount Upper Limit",
        Parse.INT,
        "CTTAUL",
    ),
    ((0x9F, 0x5D), "Available Offline Spending Amount", Parse.INT, "AOSA"),
    (
        (0x9F, 0x5E),
        "Consecutive Transaction International Upper Limit",
        Parse.INT,
        "CTIUL",
    ),
    ((0x9F, 0x6E), "Third Party Data", Parse.BYTES, None),
    (0xA5, "FCI Proprietary Template", Parse.BYTES, "FCI_PROP"),
    ((0xBF, 0x0C), "FCI Issuer Discretionary Data", Parse.BYTES, None),
]


# A list of tags which contain sensitive data, for redacting data for public display.
# This should be considered non-exhaustive and used with caution.
# Some cards may provide sensitive data with under issuer-specific tags.
SENSITIVE_TAGS = {
    0x5A,  # PAN
    (0x9F, 0x1F),  # Track1
    0x57,  # Track2
    0x56,  # Contains card number/expiry as ASCII (Mastercard prepaid)
    (0x9F, 0x6B),  # Contains card number as hex (Mastercard prepaid)
}

# European Payments Council product ID
# https://www.europeanpaymentscouncil.eu/sites/default/files/KB/files/EPC050-16%20SCS%20Volume%207%201%20-%20Bul%2001%20-%2020160229%20-%20Book%202%20-%20EEA%20Product%20Identification%20and%20usage%20in%20Selection%20of%20Application.pdf
EPC_PRODUCT_ID = {1: "Debit", 2: "Credit", 3: "Commercial", 4: "Pre-paid"}



""" This module deals with tags - identifiers of pieces of data in the EMV protocol.
    Tags can be one or more bytes long, with a continuation system a bit like UTF-8.

    Tag coding is described in:
        EMV Version 4.3 Book 3 Annex B1
"""
from functools import total_ordering
import pycountry
from .data_elements import ELEMENT_TABLE, SENSITIVE_TAGS, Parse
from ..util import format_bytes, from_hex_int, from_hex_date, decode_int

DATA_ELEMENTS = dict((tag, name) for tag, name, _, _ in ELEMENT_TABLE)
ELEMENT_FORMAT = dict(
    (tag, parse) for tag, _, parse, _ in ELEMENT_TABLE if parse is not None
)
ASCII_ELEMENTS = {tag for tag, _, parse, _ in ELEMENT_TABLE if parse == Parse.ASCII}
DOL_ELEMENTS = {tag for tag, _, parse, _ in ELEMENT_TABLE if parse == Parse.DOL}


def is_two_byte(val):
    """A tag is at least two bytes long if the least significant
    5 bits of the first byte are set."""
    return val & 0b00011111 == 0b00011111


def is_continuation(val):
    """Any subsequent byte is a continuation byte if the MSB is set."""
    return val & 0b10000000 == 0b10000000


def is_constructed(val):
    """Check if a tag represents a "constructed" value, i.e. another TLV"""
    return val & 0b00100000 == 0b00100000


def read_tag(data):
    """Read a variable-length tag from a list of bytes, starting at the
    first byte. Returns the tag, plus the number of bytes read from
    the list.

    EMV 4.3 Book 3 Annex B1
    """
    i = 0
    tag = [data[i]]
    if is_two_byte(data[i]):
        i += 1
        tag += [data[i]]
        while len(data) > i and is_continuation(data[i]):
            i += 1
            tag += [data[i]]
        i += 1
    else:
        i += 1
    return tag, i


def read_length(data):
    """Read length from a list of bytes, starting at the first byte.
    Returns the length, plus the number of bytes read from the list.

    EMV 4.3 Book 3 Annex B2
    """
    i = 0
    length = data[i]
    i += 1
    if length & 0x80:
        length_bytes_count = length & 0x7F
        length = 0
        for j in range(length_bytes_count):
            length = (length << 8) + data[i + j]
        i += length_bytes_count
    return length, i


@total_ordering
class Tag(object):
    """Represents a data tag. Provides ordering and pretty rendering."""

    def __init__(self, value):
        if type(value) in (list, tuple) and len(value) == 1:
            self.value = value[0]
        elif type(value) == int:
            self.value = value
        else:
            self.value = list(value)

    @property
    def id(self):
        if type(self.value) == list:
            return tuple(self.value)
        else:
            return self.value

    @property
    def name(self):
        if type(self.value) == list:
            return DATA_ELEMENTS.get(tuple(self.value))
        else:
            return DATA_ELEMENTS.get(self.value)

    def __hash__(self):
        if type(self.value) == list:
            return hash(tuple(self.value))
        return hash(self.value)

    def __eq__(self, other):
        if type(other) == Tag:
            return self.value == other.value
        if type(other) in (tuple, list) and type(self.value) == list:
            return tuple(other) == tuple(self.value)

        return self.name == other or self.value == other

    def __lt__(self, other):
        if type(other) == Tag:
            return self.value < other.value

        return self.value < other

    def __repr__(self):
        if type(self.value) == list:
            val = format_bytes(self.value)
        else:
            val = "%02X" % self.value

        if self.name:
            return "(%s) %s" % (val, self.name)
        else:
            return val


# Set element shortnames as static attributes on the Tag object.
for tag, _, _, shortname in ELEMENT_TABLE:
    if shortname is not None:
        setattr(Tag, shortname, tag)


def render_element(tag, value, redact=False):
    if type(tag) == Tag:
        tag = tag.id

    if redact and tag in SENSITIVE_TAGS:
        return "[REDACTED]"

    if type(value).__name__ in ("TLV", "DOL", "TagList", "ASRPD", "CVMList", "AUC"):
        return repr(value)

    if value is None:
        return "None"

    if type(value) is list and (len(value) == 0 or type(value[0]) is not int):
        return ",\n".join(render_element(tag, val, redact) for val in value)

    parse = ELEMENT_FORMAT.get(tag)
    if parse is None:
        return format_bytes(value)
    if parse == Parse.ASCII:
        return '"' + "".join(map(chr, value)) + '"'
    if parse == Parse.DEC:
        return str(from_hex_int(value))
    if parse == Parse.DATE:
        return from_hex_date(value)
    if parse == Parse.INT:
        return str(decode_int(value))
    if parse == Parse.COUNTRY:
        try:
            return pycountry.countries.get(numeric=str(from_hex_int(value))).alpha_2
        except ValueError:
            return "<Invalid country: %s>" % value
    if parse == Parse.CURRENCY:
        return pycountry.currencies.get(numeric=str(from_hex_int(value))).alpha_3
    return format_bytes(value)



from setuptools import setup
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open("README.md") as f:
    long_description = f.read()

version = {}
with open(path.join(here, "emv", "__init__.py")) as fp:
    exec(fp.read(), version)

setup(
    name="emv",
    version=version["__version__"],
    description="EMV Smartcard Protocol Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    author="Russ Garrett",
    author_email="russ@garrett.co.uk",
    url="https://github.com/russss/python-emv",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    keywords="smartcard emv payment",
    python_requires=">=3.4",
    packages=["emv", "emv.protocol", "emv.command"],
    install_requires=[
        "enum-compat==0.0.3",
        "argparse==1.4.0",
        "pyscard==2.0.0",
        "pycountry==20.7.3",
        "terminaltables==3.1.0",
        "click==7.1.2",
    ],
    entry_points={"console_scripts": {"emvtool=emv.command.client:run"}},
)
